 /* -LICENSE-START-
 ** Copyright (c) 2024 Blackmagic Design
 **  
 ** Permission is hereby granted, free of charge, to any person or organization 
 ** obtaining a copy of the software and accompanying documentation (the 
 ** "Software") to use, reproduce, display, distribute, sub-license, execute, 
 ** and transmit the Software, and to prepare derivative works of the Software, 
 ** and to permit third-parties to whom the Software is furnished to do so, in 
 ** accordance with:
 ** 
 ** (1) if the Software is obtained from Blackmagic Design, the End User License 
 ** Agreement for the Software Development Kit (“EULA”) available at 
 ** https://www.blackmagicdesign.com/EULA/DeckLinkSDK; or
 ** 
 ** (2) if the Software is obtained from any third party, such licensing terms 
 ** as notified by that third party,
 ** 
 ** and all subject to the following:
 ** 
 ** (3) the copyright notices in the Software and this entire statement, 
 ** including the above license grant, this restriction and the following 
 ** disclaimer, must be included in all copies of the Software, in whole or in 
 ** part, and all derivative works of the Software, unless such copies or 
 ** derivative works are solely in the form of machine-executable object code 
 ** generated by a source language processor.
 ** 
 ** (4) THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 ** OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 ** FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT 
 ** SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE 
 ** FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE, 
 ** ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 ** DEALINGS IN THE SOFTWARE.
 ** 
 ** A copy of the Software is available free of charge at 
 ** https://www.blackmagicdesign.com/desktopvideo_sdk under the EULA.
 ** 
 ** -LICENSE-END-
 */


#include "platform.h"

#include <iomanip>
#include <iostream>

#include "com_ptr.h"

com_ptr<IDeckLink> GetDeckLinkByPersistentID(std::string persistentID);
com_ptr<IDeckLinkIPFlow> GetDeckLinkIPFlow(com_ptr<IDeckLink>& deckLink, BMDIPFlowDirection direction, BMDIPFlowType type);
void Connect(com_ptr<IDeckLinkIPFlow>& source, com_ptr<IDeckLinkIPFlow>& dest);
void PrintUsage();

int main(int argc, const char * argv[])
{
	Initialize();

	if (argc != 3)
	{
		PrintUsage();
		return 1;
	}
	
	try {
		
		com_ptr<IDeckLink> source = GetDeckLinkByPersistentID(argv[1]);
		com_ptr<IDeckLink> dest = GetDeckLinkByPersistentID(argv[2]);

		{
			com_ptr<IDeckLinkIPFlow> sourceVideo = GetDeckLinkIPFlow(source, bmdDeckLinkIPFlowDirectionOutput, bmdDeckLinkIPFlowTypeVideo);
			com_ptr<IDeckLinkIPFlow> sourceAudio = GetDeckLinkIPFlow(source, bmdDeckLinkIPFlowDirectionOutput, bmdDeckLinkIPFlowTypeAudio);
			com_ptr<IDeckLinkIPFlow> sourceAncillary = GetDeckLinkIPFlow(source, bmdDeckLinkIPFlowDirectionOutput, bmdDeckLinkIPFlowTypeAncillary);
			
			com_ptr<IDeckLinkIPFlow> destVideo = GetDeckLinkIPFlow(dest, bmdDeckLinkIPFlowDirectionInput, bmdDeckLinkIPFlowTypeVideo);
			com_ptr<IDeckLinkIPFlow> destAudio = GetDeckLinkIPFlow(dest, bmdDeckLinkIPFlowDirectionInput, bmdDeckLinkIPFlowTypeAudio);
			com_ptr<IDeckLinkIPFlow> destAncillary = GetDeckLinkIPFlow(dest, bmdDeckLinkIPFlowDirectionInput, bmdDeckLinkIPFlowTypeAncillary);

			Connect(sourceVideo, destVideo);
			Connect(sourceAudio, destAudio);
			Connect(sourceAncillary, destAncillary);
		}
		
		std::cout << "3 source flows were connected to 3 destination flows" << std::endl;
		
		return 0;
	}
	catch (const std::runtime_error& error)
	{
		std::cerr << "There were errors when running this example:" << std::endl;
		std::cerr << "  " << error.what() << std::endl;
		return 1;
	}
}

com_ptr<IDeckLink> GetDeckLinkByPersistentID(std::string persistentID)
{
	com_ptr<IDeckLinkIterator> deckLinkIterator;
	if (GetDeckLinkIterator(deckLinkIterator.releaseAndGetAddressOf()) != S_OK)
		throw std::runtime_error("Could not get IDeckLinkIterator interface");

	const uint32_t persistentIDAsUInt = (uint32_t)std::stoull(persistentID, nullptr, 16);
	
	while (true)
	{
		com_ptr<IDeckLink> deckLink;
		if (deckLinkIterator->Next(deckLink.releaseAndGetAddressOf()) != S_OK)
			break;

		com_ptr<IDeckLinkProfileAttributes> deckLinkProfileAttributes;
		if (deckLink->QueryInterface(IID_IDeckLinkProfileAttributes, (void**)deckLinkProfileAttributes.releaseAndGetAddressOf()) != S_OK)
			throw std::runtime_error("Could not get IDeckLinkProfileAttributes interface for DeckLink");

		INT64_SIGNED thisPersistentID = 0;
		if (deckLinkProfileAttributes->GetInt(BMDDeckLinkPersistentID, &thisPersistentID) != S_OK)
			throw std::runtime_error("Could not get PersistentID interface for DeckLink");

		if (static_cast<uint32_t>(thisPersistentID) == persistentIDAsUInt)
			return deckLink;
	}

	throw std::runtime_error("Could not get DeckLink matching persistent ID " + persistentID);
}

com_ptr<IDeckLinkIPFlow> GetDeckLinkIPFlow(com_ptr<IDeckLink>& deckLink, BMDIPFlowDirection direction, BMDIPFlowType type)
{
	com_ptr<IDeckLinkIPExtensions> deckLinkIPExtensions;
	if (deckLink->QueryInterface(IID_IDeckLinkIPExtensions, (void**)deckLinkIPExtensions.releaseAndGetAddressOf()) != S_OK)
		throw std::runtime_error("Could not get IDeckLinkIPExtensions interface for DeckLink");

	com_ptr<IDeckLinkIPFlowIterator> iterator;
	if (deckLinkIPExtensions->GetDeckLinkIPFlowIterator(iterator.releaseAndGetAddressOf()) != S_OK)
		throw std::runtime_error("Could not get IDeckLinkIPFlowIterator interface for DeckLink");

	com_ptr<IDeckLinkIPFlow> candidate;
	
	while (iterator->Next(candidate.releaseAndGetAddressOf()) == S_OK)
	{
		com_ptr<IDeckLinkIPFlowAttributes> deckLinkFlowAttributes;
		if (candidate->QueryInterface(IID_IDeckLinkIPFlowAttributes, (void**)deckLinkFlowAttributes.releaseAndGetAddressOf()) != S_OK)
			throw std::runtime_error("Could not get IDeckLinkIPFlowAttributes interface for Flow");
		
		const bool match = [&]()
		{
			INT64_SIGNED candidateType;
			
			if (deckLinkFlowAttributes->GetInt(bmdDeckLinkIPFlowType, &candidateType) != S_OK)
				return false;
			
			if ((BMDIPFlowType)candidateType != type)
				return false;
			
			INT64_SIGNED candidateDirection;
			
			if (deckLinkFlowAttributes->GetInt(bmdDeckLinkIPFlowDirection, &candidateDirection) != S_OK)
				return false;
			
			if ((BMDIPFlowDirection)candidateDirection != direction)
				return false;
			
			return true;
		}();
		
		if (match)
			return candidate;
	}

	throw std::runtime_error("Could not find Flow");
}

void Connect(com_ptr<IDeckLinkIPFlow>& source, com_ptr<IDeckLinkIPFlow>& dest)
{
	dest->Disable();
	
	// Fetch the SDP from the source

	com_ptr<IDeckLinkIPFlowStatus> sourceStatus;
	if (source->QueryInterface(IID_IDeckLinkIPFlowStatus, (void**)sourceStatus.releaseAndGetAddressOf()) != S_OK)
		throw std::runtime_error("Could not get IDeckLinkIPFlowStatus interface for Flow");
	
	STRINGOBJ sdp;
	
	if (sourceStatus->GetString(bmdDeckLinkIPFlowSDP, &sdp) != S_OK)
		throw std::runtime_error("Could not fetch SDP from source Flow");

	// Set the peer SDP in the destination

	com_ptr<IDeckLinkIPFlowSetting> destSetting;
	if (dest->QueryInterface(IID_IDeckLinkIPFlowSetting, (void**)destSetting.releaseAndGetAddressOf()) != S_OK)
	{
		STRINGFREE(sdp);
		throw std::runtime_error("Could not get IDeckLinkIPFlowSetting interface for Flow");
	}
	
	if (destSetting->SetString(bmdDeckLinkIPFlowPeerSDP, sdp) != S_OK)
	{
		STRINGFREE(sdp);
		throw std::runtime_error("Could not set SDP for dest Flow");
	}
	
	std::string value;
	StringToStdString(sdp, value);
	
	std::cout << value << std::endl;
	
	STRINGFREE(sdp);

	dest->Enable();

	source->Enable();
}

void PrintUsage()
{
	std::cout << "Usage: DeckLinkIPExample [sourcePersistentID] [destPersistentID]" << std::endl << std::endl;
	
	std::cout << "Connects all flows from [sourcePersistentID] to [destPersistentID]" << std::endl << std::endl;
	
	std::cout << "Candidate devices:" << std::endl << std::endl;
	
	std::cout << "Persistent ID  Name" << std::endl;
	std::cout << "=============  ====" << std::endl;
	
	com_ptr<IDeckLinkIterator> deckLinkIterator;
	if (GetDeckLinkIterator(deckLinkIterator.releaseAndGetAddressOf()) != S_OK)
		throw std::runtime_error("Could not get IDeckLinkIterator interface");
	
	while (true)
	{
		com_ptr<IDeckLink> deckLink;
		if (deckLinkIterator->Next(deckLink.releaseAndGetAddressOf()) != S_OK)
			break;
			
		STRINGOBJ deviceNameRaw;
		if (deckLink->GetDisplayName(&deviceNameRaw) != S_OK)
			throw std::runtime_error("Could not get DisplayName for DeckLink");

		std::string deviceName;
		StringToStdString(deviceNameRaw, deviceName);
		
		STRINGFREE(deviceNameRaw);
		
		com_ptr<IDeckLinkProfileAttributes> deckLinkProfileAttributes;
		if (deckLink->QueryInterface(IID_IDeckLinkProfileAttributes, (void**)deckLinkProfileAttributes.releaseAndGetAddressOf()) != S_OK)
			throw std::runtime_error("Could not get IDeckLinkProfileAttributes interface for DeckLink");

		INT64_SIGNED persistentID = 0;
		if (deckLinkProfileAttributes->GetInt(BMDDeckLinkPersistentID, &persistentID) != S_OK)
			throw std::runtime_error("Could not get PersistentID interface for DeckLink");

		std::cout << std::hex << std::setw(8) << persistentID << "       " << deviceName << std::endl;
	}
}
